package org.naokishibata.generic;

import java.util.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.swing.*;

public class AnimationPanel implements WindowListener {
    private int width, height;

    private Window window;
    private VolatileImage vFrontBuffer = null, vBackBuffer = null;
    private BufferStrategy bufferStrategy = null;

    private Graphics2D gf = null, gb = null;

    private Color backgroundColor = Color.white;

    private Frame frame = null;
    private VPanel panel = null;

    private boolean exitWhenClosed = false;
    private boolean enableBufferStrategy = false;

    public AnimationPanel(int width, int height, boolean ebs) {
	this.width = width;
	this.height = height;
	this.enableBufferStrategy = ebs;

	this.window = null;
	vFrontBuffer = newVolatileImage(width, height);
	vBackBuffer = newVolatileImage(width, height);

	gf = vFrontBuffer.createGraphics();
	gb = vBackBuffer.createGraphics();
    }

    public AnimationPanel(int width, int height) { this(width, height, false); }

    public AnimationPanel(Window w, int width, int height, boolean ebs) {
	this.width = width;
	this.height = height;
	this.enableBufferStrategy = ebs;

	this.window = w;
	vFrontBuffer = newVolatileImage(width, height);
	vBackBuffer = newVolatileImage(width, height);

	if (window != null && enableBufferStrategy) {
	    try {
		window.createBufferStrategy(2);
		bufferStrategy = window.getBufferStrategy();
	    } catch(IllegalArgumentException ex) {
		Debug.abort(ex);
	    } catch(IllegalStateException ex) {
	    }
	}

	gf = vFrontBuffer.createGraphics();
	gb = vBackBuffer.createGraphics();
    }

    public AnimationPanel(Window w, int width, int height) { this(w, width, height, false); }

    private class VPanel extends JPanel implements ComponentListener {
	public VPanel() {
	    super();
	    //super.setMinimumSize(new Dimension(1,1));
	    //super.setMaximumSize(new Dimension(2048,2048));
	    addComponentListener(this);
	}

	//public Dimension getPreferredSize() { return new Dimension(width, height); }

	public void paint(Graphics g) {
	    super.paint(g);

	    Image im = vBackBuffer;
	    synchronized(im) {
		g.drawImage(im, 0, 0, null);
	    }
	}

	public void update(Graphics g) { paint(g); }

	public void componentHidden(ComponentEvent e) {}
	public void componentMoved(ComponentEvent e) {}
	public void componentShown(ComponentEvent e) {}

	public void componentResized(ComponentEvent e) {
	    Rectangle r = getBounds();
	    setSize0(r.width, r.height);
	}
    }

    public JPanel makePanel() {
	if (panel != null) return panel;
	panel = new VPanel();
	return panel;
    }

    public JFrame makeFrame(String name) {
	if (frame != null) {
	    Debug.abort("Already has a frame");
	}

	panel = (VPanel)makePanel();

        frame = new JFrame(name);

        frame.setSize(1, 1);
        frame.setResizable(true);
        ((JFrame)frame).setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        frame.addWindowListener(this);

        ((JFrame)frame).getContentPane().add("Center", panel);

        AWTUtils.placeAtTheCenter(frame);

        frame.pack();
        frame.setVisible(true);

        while(frame.getWidth() == 1 && frame.getHeight() == 1) {
            Insets in = frame.getInsets();
            frame.setSize(width + in.left + in.right, height + in.top  + in.bottom);

            try { Thread.sleep(100); } catch(InterruptedException e) {}
        }

        AWTUtils.placeAtTheCenter(frame);

	window = frame;

	if (enableBufferStrategy) {
	    try {
		window.createBufferStrategy(2);
		bufferStrategy = window.getBufferStrategy();
	    } catch(IllegalArgumentException ex) {
		Debug.abort(ex);
	    } catch(IllegalStateException ex) {
	    }
	}

        return (JFrame)frame;
    }

    public JFrame makeFrame() { return makeFrame(MiscUtils.getMainClassName()); }

    /*
    public Frame makeFullScreenFrame() {
	if (frame != null) {
	    Debug.abort("Already has a frame");
	}

	panel = (VPanel)makePanel();

        window = frame = new FullScreenFrame();
	enableBufferStrategy = true;

	try {
	    window.createBufferStrategy(2);
	    bufferStrategy = window.getBufferStrategy();
	} catch(IllegalArgumentException ex) {
	    Debug.abort(ex);
	} catch(IllegalStateException ex) {
	}

	frame.setVisible(true);

	Rectangle r = frame.getBounds();
	setSize(r.width, r.height);

        return frame;
    }
    */

    public Frame frame() { return frame; }

    public void setVisible(boolean b) {}
    public void setIconifyWhenClosed(boolean b) {}
    public void setExitWhenClosed(boolean b) { exitWhenClosed = b; }
    public void windowClosing(WindowEvent e) { if (exitWhenClosed) System.exit(0); }

    public DisplayMode[] getDisplayModes() {
	GraphicsDevice gd = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
	return gd.getDisplayModes();	
    }

    public void render(Graphics g) {
	Image im = vBackBuffer;
	synchronized(im) {
	    g.drawImage(im, 0, 0, null);
	}
    }

    public int getWidth() { return width; }
    public int getHeight() { return height; }

    public void setSize(int width, int height) {
	this.width = width;
	this.height = height;
    }

    private void setSize0(int width, int height) { setSize(width, height); }

    public static VolatileImage newVolatileImage(int width, int height) {
	GraphicsConfiguration gc = GraphicsEnvironment.getLocalGraphicsEnvironment().
	    getDefaultScreenDevice().getDefaultConfiguration();
	VolatileImage v = gc.createCompatibleVolatileImage(width, height);
	v.setAccelerationPriority(1);

	return v;
    }

    public Graphics2D getGraphics2D() {
	if (bufferStrategy != null) {
	    if (gf != null) gf.dispose();
	    gf = (Graphics2D)bufferStrategy.getDrawGraphics();

	    if (frame != null) {
		Insets in = frame.getInsets();
		gf.setTransform(AffineTransform.getTranslateInstance(in.left, in.top));
	    }

	    gf.setColor(backgroundColor);
	    gf.fillRect(0, 0, vFrontBuffer.getWidth(), vFrontBuffer.getHeight());

	    return gf;
	}

	GraphicsConfiguration gc = GraphicsEnvironment.getLocalGraphicsEnvironment().
	    getDefaultScreenDevice().getDefaultConfiguration();

	if (vFrontBuffer.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE) {
	    System.out.println("incompatible");
	    vFrontBuffer = newVolatileImage(width, height);
	    gf = vFrontBuffer.createGraphics();
	}

	gf.setColor(backgroundColor);
	gf.fillRect(0, 0, vFrontBuffer.getWidth(), vFrontBuffer.getHeight());

	return gf;
    }

    public void swapBuffer() {
	if (bufferStrategy != null) {
	    bufferStrategy.show();
	    return;
	}

	synchronized(vBackBuffer) {
	    VolatileImage im = vFrontBuffer;
	    vFrontBuffer = vBackBuffer;
	    vBackBuffer = im;
	}

	Graphics2D g = gf;
	gf = gb;
	gb = g;

	if (vFrontBuffer.getWidth() < width || vFrontBuffer.getHeight() < height) {
	    vFrontBuffer = newVolatileImage(width, height);
	    gf.dispose();
	    gf = vFrontBuffer.createGraphics();
	}

	if (panel != null) panel.repaint();

	if (window != null && enableBufferStrategy) {
	    try {
		window.createBufferStrategy(2);
		bufferStrategy = window.getBufferStrategy();
	    } catch(IllegalArgumentException ex) {
		Debug.abort(ex);
	    } catch(IllegalStateException ex) {
	    }
	}
    }

    public void setBackgroundColor(Color c) { backgroundColor = c; }

    //

    public void windowOpened(WindowEvent e) {}
    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
}
